#!/usr/bin/env python

__version__ = '$Revision: 4853 $'.split()[1]
__date__ = '$Date: 2006-10-03 11:44:03 -0400 (Tue, 03 Oct 2006) $'.split()[1]
__author__ = 'Kurt Schwehr'

__doc__='''
Read an arg xml definition and produce code for C++ file that uses gengetopt to read the options.

@note: ET is an abreviation for ElementTree.  Generally using the ElementTree API as provided by lxml

@requires: U{lxml<http://codespeak.net/lxml/>}

@author: '''+__author__+'''
@version: ''' + __version__ +'''
@copyright: 2006
@var __date__: Date of last svn commit
@undocumented: __version__ __author__ __doc__ myparser 
@since: 2006-Sep-11
@status: under development
@organization: U{CCOM<http://ccom.unh.edu/>}

@bug: FIX: probably better to make a class from each option, no?
@bug: FIX: Needs a serious cleanup!
@bug: Need to make sure that there are not xml troubles in the encoded text for save
@bug: Should probably use libxml2 for the saving!

@license: LGPL
'''

from genargs.option import Option


######################################################################
def writeLibxml2ReaderHelpers(out):
    '''
    Functions to make dealing with libxml2 manageable.
    
    @param out: Open file to write static libxml2 heards
    @todo: only write out the static functions that are used
    '''
    o=out

    o.write('''
//////////////////////////////////////////////////////////////////////
// HELPER section for libxml2
//////////////////////////////////////////////////////////////////////
#include <libxml/parser.h>
#include <libxml/xpath.h>
#include <libxml/tree.h>

#include <iostream>
#include <string>
#include <vector>

static inline bool xmlStreq(const xmlChar *a,const xmlChar *b) {
    return ( (0==xmlStrcmp(a,b)) ? true:false ) ;
}

static inline bool xmlStreq(const std::string &a,const xmlChar *b) {
    return ( (0==xmlStrcmp((const xmlChar *)a.c_str(),b)) ? true:false ) ;
}

static inline bool xmlStreq(const xmlChar *a,const std::string &b) {
    return ( (0==xmlStrcmp(a,(const xmlChar *)b.c_str())) ? true:false ) ;
}

/// Tired of this C api.
#define XPATHFREE xmlXPathFreeContext(context); xmlXPathFreeObject(result);

/**
   Returns the text inside the node converted to a boolean

   @return: sets found to false
   @bug: only hands '1' or '0'
 */
static bool getBool(xmlDocPtr doc,std::string xpath, bool &found) {
    found = true; // Innocent until proven guilty
    //std::cout << "getBool xpath: " << xpath << "\\n";

    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);
    if(xmlXPathNodeSetIsEmpty(result->nodesetval)) { found=false; XPATHFREE; return false; }

    xmlNodePtr node = result->nodesetval->nodeTab[0];
    if (!node->xmlChildrenNode) { found=false; XPATHFREE; return false; }

    xmlChar *strValue = node->xmlChildrenNode->content;
    if (!strValue || !xmlStreq(node->xmlChildrenNode->name,std::string("text"))) 
	{ found=false; XPATHFREE; return false; }

    const int intVal = atoi((const char *)strValue);
    if (0!=intVal and 1!=intVal) {found=false; XPATHFREE; return false;}
    XPATHFREE;
    
    return bool(intVal);
}

/**
   Returns the text inside the node converted to a int

   @return: sets found to false

 */
static int getInt(xmlDocPtr doc,std::string xpath, bool &found) {
    found = true; // Innocent until proven guilty

    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);
    if(xmlXPathNodeSetIsEmpty(result->nodesetval)) { found=false; XPATHFREE; return false; }

    xmlNodePtr node = result->nodesetval->nodeTab[0];
    if (!node->xmlChildrenNode) { found=false; XPATHFREE; return false; }

    xmlChar *strValue = node->xmlChildrenNode->content;
    if (!strValue || !xmlStreq(node->xmlChildrenNode->name,std::string("text"))) 
	{ found=false; XPATHFREE; return 0; }

    const float intVal = atoi((const char *)strValue);
    XPATHFREE;
    
    return intVal;
}

/**
   Returns the text inside the node converted to a float

   @return: sets found to false

 */
static float getFloat(xmlDocPtr doc,std::string xpath, bool &found) {
    found = true; // Innocent until proven guilty

    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);
    if(xmlXPathNodeSetIsEmpty(result->nodesetval)) { found=false; XPATHFREE; return false; }

    xmlNodePtr node = result->nodesetval->nodeTab[0];
    if (!node->xmlChildrenNode) { found=false; XPATHFREE; return false; }

    xmlChar *strValue = node->xmlChildrenNode->content;
    if (!strValue || !xmlStreq(node->xmlChildrenNode->name,std::string("text"))) 
	{ found=false; XPATHFREE; return 0.; }

    const float floatVal = (strValue?atof((const char *)strValue):0.);
    XPATHFREE;
    
    return floatVal;
}


/**
   Get the text within a node.  Does not deal with nodeTab being in random order if multiple hits

   @return: NOT FOUND string on failure and sets found to false

 */
static std::string getString(xmlDocPtr doc,std::string xpath, bool &found) {
    found = true; // Innocent until proven guilty
    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);

    if(xmlXPathNodeSetIsEmpty(result->nodesetval)){ found=false; XPATHFREE; return std::string("NOT FOUND"); }

    xmlNodePtr node = result->nodesetval->nodeTab[0];
    if (!node->xmlChildrenNode) {found=false; XPATHFREE; return std::string("NOT FOUND"); }

    xmlChar *strValue = node->xmlChildrenNode->content;
    if (!xmlStreq(node->xmlChildrenNode->name,std::string("text"))) strValue=0;

    const std::string resultStr= (strValue?(const char *)strValue:"NOT FOUND");
    if (!strValue) {found=false;}
    XPATHFREE;
    return resultStr;
}

''')


    o.write('''
////////////////////////////////////////
// ARRAY ACCESSORS
////////////////////////////////////////
static std::vector<int> getIntArray(xmlDocPtr doc,std::string xpath, bool &found) {
    found = true;
    std::vector<int> v;
    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);

    if(xmlXPathNodeSetIsEmpty(result->nodesetval)){ found=false; XPATHFREE; return v; }

    xmlNodePtr node = result->nodesetval->nodeTab[0];
    if (!node->xmlChildrenNode) {found=false; XPATHFREE; return v; }

    // FIX: do we warn on bad values?
    for (xmlNodePtr valueNode=node->xmlChildrenNode; valueNode; valueNode=valueNode->next) {
	if (!xmlStreq(valueNode->name,"value")) continue;
	xmlChar *strValue = valueNode->xmlChildrenNode->content;
	if (!strValue || !xmlStreq(valueNode->xmlChildrenNode->name,std::string("text")) ) continue;
	v.push_back(atoi((const char *)strValue));
    }
    
    XPATHFREE;
    return v;
}


static std::vector<float> getFloatArray(xmlDocPtr doc,std::string xpath, bool &found) {
    found = true;
    std::vector<float> v;
    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);

    if(xmlXPathNodeSetIsEmpty(result->nodesetval)){ found=false; XPATHFREE; return v; }

    xmlNodePtr node = result->nodesetval->nodeTab[0];
    if (!node->xmlChildrenNode) {found=false; XPATHFREE; return v; }

    // FIX: do we warn on bad values?
    for (xmlNodePtr valueNode=node->xmlChildrenNode; valueNode; valueNode=valueNode->next) {
	if (!xmlStreq(valueNode->name,"value")) continue;
	xmlChar *strValue = valueNode->xmlChildrenNode->content;
	if (!strValue || !xmlStreq(valueNode->xmlChildrenNode->name,std::string("text")) ) continue;
	v.push_back(atof((const char *)strValue));
    }
    
    XPATHFREE;
    return v;
}

static std::vector<std::string> getStringArray(xmlDocPtr doc,std::string xpath, bool &found) {
    found = true;
    std::vector<std::string> vs;
    xmlXPathContextPtr context = xmlXPathNewContext(doc);
    xmlXPathObjectPtr result = xmlXPathEvalExpression((xmlChar *)xpath.c_str(), context);

    if(xmlXPathNodeSetIsEmpty(result->nodesetval)){ found=false; XPATHFREE; return vs; }

    xmlNodePtr node = result->nodesetval->nodeTab[0];
    if (!node->xmlChildrenNode) {found=false; XPATHFREE; return vs; }

    // FIX: do we warn on bad values?
    for (xmlNodePtr valueNode=node->xmlChildrenNode; valueNode; valueNode=valueNode->next) {
	if (!xmlStreq(valueNode->name,"value")) continue;
	xmlChar *strValue = valueNode->xmlChildrenNode->content;
	if (!strValue || !xmlStreq(valueNode->xmlChildrenNode->name,std::string("text")) ) continue;
	vs.push_back((const char *)strValue);
    }
    
    XPATHFREE;
    return vs;
}

''')

    return

def writeLibxml2Reader(out,root,indent='    ',className='Parameters',verbose=False):
    '''
    Spit out C++ method to read in variables from an argdump file.
    
    @param out: open C++ source file to write to.
    @param root: Element Tree root node
    @param indent: how much to ident by.
    @param className: c++ class that we are writing the method for
    @param verbose: set to true for more to stdout.
    '''

    o = out
    o.write('bool '+className+'::load(const std::string &filename) {\n')
    o.write('''
    xmlDocPtr doc = xmlParseFile(filename.c_str());
    if (!doc) return false;
    bool found;

''')

    for category in root:
	if not category.tag:
	    continue # skip comments
	o.write('\n')
	o.write(indent+'////////////////////////////////////////\n')
	o.write(indent+'//  Catagory: '+category.attrib['name']+'\n')
	o.write(indent+'////////////////////////////////////////\n\n')
	for optET in category:
	    if optET.tag!='opt': continue # Skip comments and other stuff
	    if verbose: print '    opt:', optET.attrib['name']
	    opt = Option(optET)
	    t = opt['type']
	    print opt['name'],'type:',opt['type']
	    #o.write(indent+'std::cout << "'+indent+'loading '+opt['name']+'\\n";\n')
	    
	    # craft xpath - FIX: how careful/exact should this be?
	    xpathStr='/argdump[@className=\''+className+'\']/optvalues[@name=\''+opt['name']+'\']'

	    o.write(indent+'const '+opt['cxxType']+' tmp_'+opt['variable']+'=get')
	    o.write(t.capitalize())
	    if opt['repeatable']: o.write('Array')
	    o.write('(doc,"'+xpathStr+'",found);\n')

	    o.write(indent+'if (found) '+opt['variable']+'='+'tmp_'+opt['variable']+';\n')
	    #o.write(indent+'else std::cout << "Did not find '+xpathStr+'\\n";\n')


    o.write('\n'+indent+'return true; // FIX: what constitutes failure?\n')
    o.write(' } // '+className+'::load(const std::string &filename)\n\n')

    return

######################################################################
def writeOptHeader(out,optET,indent='    '
		   ,publicMembers=True
		   ,includeGetSet=True
		   ):
    '''
    Handle one line of the option file

    @param out: is a file object that will be written to
    @param optET: is an element tree node containing an opt.
    '''

    opt = Option(optET)
    var = opt['variable']
    capVar=var.capitalize()
    varType=opt['type']

    if varType=='string': varType='std::string'

    o = out
    o.write('\n'+indent+'////////////////////\n');
    o.write(indent+'''/** @defgroup '''+var+' '+opt['shortDescription']+'\n') 
    o.write(opt['longDescription'])
    o.write('\n'+indent+'@{\n')
    o.write(indent+'*/\n')

    o.write('public:\n')
    if includeGetSet:
	o.write(indent+'inline '+varType+' get'+capVar)
	if opt['repeatable']:
	    o.write('(const size_t index) const { return '+var+'[index]; } ')
	else:
	    o.write('() const { return '+var+'; }')
	    o.write('///< retrieve the param\n')

	o.write(indent+'inline void set'+capVar+'(const '+varType+' &newVal')
	if opt['repeatable']:
	    o.write(',const size_t index) { '+var+'[index]')
	else:
	    o.write(') { '+var+' ')
	o.write(' = newVal; } ///< set the param\n')

    # Array only operations
    if opt['repeatable']:
	o.write(indent+'inline void append'+capVar+'(const '+varType+' &newVal) {'+var+'.push_back(newVal);} ///< Add an entry on the end\n')
	o.write(indent+'inline void    pop'+capVar+'() {'+var+'.pop_back();} ///< Remove last entry\n')
	o.write(indent+'inline void   size'+capVar+'() {'+var+'.size();} ///< Length of the array\n')

    if not publicMembers: o.write('private:\n')
    
    o.write(indent+'/*!\n')
    o.write(indent+' @brief '+opt['shortDescription']+'\n')
    o.write(indent+' \n')

    o.write(indent+' Attributes:\n')
    o.write(indent+'   - Name: '+opt['name']+'\n')
    o.write(indent+'   - Short Option:'+opt['shortOpt']+'\n')
    o.write(indent+'   - Long Option:'+opt['longOpt']+'\n')
    o.write(indent+'   - Default: '+str(opt['default'])+'\n')
    o.write(indent+'   - Units: '+str(opt['units'])+'\n')
    o.write(indent+' \n')

    o.write(indent+' \n')
    
    o.write(indent+' '+opt['longDescription']+'\n')
    o.write(indent+' */\n')

    if opt['repeatable']:
	o.write(indent+opt['cxxType']+' '+var+';\n')
    else:
	o.write(indent+varType+' '+var+';\n')
    
    o.write(indent+'/** @} */ // End of '+optET.attrib['variable']+' group\n')



def writeCxxHeader(argsXmlFilename='genargs.xml', outFilename='Parameters.h', 
		   gengetoptStructName='gengetopt_args_info', 
		   className='Parameters',indent='    '
		   ,publicMembers=True
		   ,includeGetSet=True
		   ,verbose=False):
    '''
    FIX: write docs here

    @param argsXmlFilename: FIX: write
    @param outFilename: FIX: write
    @param gengetoptStructName: FIX: write
    @param className: FIX: write
    @param indent: FIX: write
    '''
    from lxml import etree as ET
    root = ET.parse(argsXmlFilename).getroot()

    o = file(outFilename,'w')
    o.write('''// -*- c++ -*- tell emacs what this is C++.  Important if .h is used for the file name
/** @file
    @brief Autogenerated source file defining the parameter class
 */

#ifndef _'''+className.upper()+'''_H_ // include guard
#define _'''+className.upper()+'''_H_

#include <string>
#include <vector>

/**
  @brief Autogenerated parameter container

  This class contains the parameters needed for a range of
  applications.  The file is AUTOGENERATED.  Do NOT edit.
*/
class '''+className+''' {
public:
   /** @brief parameter class that allows for command line options and config files.

   All is used for writing all values.
   Set means that variable was set by one of a set method
   */

'''+indent+className+'''(void) {};  ///< Initialize without defaults
'''+indent+className+'''(const bool useDefaults); ///< Initialize with/without the defaults
'''+indent+className+'''(const int argc, char * const *argv);  ///< Initialize from command line args
'''+indent+'~'+className+'''(void) {};  ///< Initialize without defaults.  FIX: does this need more?

'''+indent+'''void init(); ///< Set the class to an initial state of unknown
'''+indent+'''void defaults(); ///< Set the class to an initial state

   // FIX: implement these two
'''+indent+'''bool save(const std::string &filename) const; ///< Write xml of parameters to a file, but use only the listed source
'''+indent+'''bool load(const std::string &filename); ///< Read parameters from an xml file

'''+indent+'''void print(); ///< Dump to stdout


''')

#'''+indent+'''bool save(const std::string &filename, const Source &source=All) const; 
#'''+indent+'''enum Source { Invalid, Defaults, File, Set, Unknown, All };


    for category in root:
	if not category.tag:
	    continue # skip comments
	o.write('\n')
	o.write(indent+'////////////////////////////////////////\n')
	o.write(indent+'//  Catagory: '+category.attrib['name']+'\n')
	o.write(indent+'////////////////////////////////////////\n\n')
	for opt in category:
	    if opt.tag!='opt': continue # Skip comments and other stuff
	    if verbose: print '    opt:', opt.attrib['name']
	    writeOptHeader(o,opt,indent,publicMembers,includeGetSet)

    o.write('}; // class '+className+'\n')
    o.write('#endif // _'+className.upper()+'_H_ // include guard\n')

######################################################################
######################################################################

def writeOptInit(out,optET,indent='    '):
    '''
    Handle one line of the option file for initialization to the default

    @param out: is a file object that will be written to
    @param opt: is an element tree node containing an opt.
    '''
    opt = Option(optET)
    var = opt['variable']
    varType = opt['cxxType']
    #varType = opt.attrib['type']

    default=opt['default']

    # FIX: how to handle array defaults?
    if None==default or opt['repeatable']:
	return

    #if varType=='string': varType='std::string'

    # FIX: what to do about multiple/repeaters?  They need to have vectors

    print 'init: vartype=',varType
    if varType=='bool': default=default.lower()
    if varType=='string' or varType=='std::string':
	if len(default)==0: return # String intialize themselves to emtpy
	default=opt['cxxType']+'("'+default+'")'
    #if 'none'!=default.lower():
    out.write(indent+var+'='+default+';\n')

######################################################################

def cmdlineConstructor(o,root,className,indent,gengetoptStructName):
    '''
    Build the bridge from the gengetopt (ggo) structure to the c++ struct
    @param o: Open file to write to
    @param root: elementtree for the args
    @param className: string giving the name of the C++ class
    @param gengetoptStructName: name of the structure in C by ggo
    '''
    o.write(className+'::'+className+'(const int argc, char * const *argv) {\n')
    # /* FIX: is an init needed here?'''+indent+'''init(); */
    o.write(indent+gengetoptStructName+' a; // a == arguments\n')
    o.write(indent+'int newArgc=argc;\n')
    o.write( indent+'if (0!=cmdline_parser(newArgc,argv,&a)){\n')
    o.write(2*indent+'std::cerr<<"MELT DOWN: should never get here"<<std::endl;\n')
    o.write(2*indent+'return /*(EXIT_FAILURE)*/; // FIX: how to return the failure?\n')
    o.write(indent+'}\n\n')
    #/* FIX: Copy all of the args from here */

    for category in root:
	if not category.tag:
	    continue # skip comments
	for optET in category:
	    if optET.tag!='opt': continue # Skip comments and other stuff

	    opt = Option(optET)
	    var=opt['variable']

	    if not opt['repeatable']:
		if opt['type']==bool:
		    o.write(indent+var+' = bool(a.'+opt['gengetoptVar']+');\n')
		elif 'string'==opt['type']:
		    # Make sure to check for a null pointer/ no string!
		    o.write(indent+'if (a.'+opt['gengetoptVar']+') '+var+' = a.'+opt['gengetoptVar']+';\n')
		else:
		    o.write(indent+var+' = a.'+opt['gengetoptVar']+';\n')
	    else:
		o.write(indent+'for (unsigned int i=0; i<a.'+opt['gengetoptLength']+';i++) {\n')
		o.write(indent+'    '+opt['variable']+'.push_back(a.'+opt['gengetoptVar']+'[i]);\n')
		o.write(indent+'}\n')


    o.write('} // '+className+'::'+className+'(argc,argv)\n\n')

######################################################################

def printFunc(out,root,className,indent, verbose=False):
    '''
    Create a ::print c++ function to dump the class state to stdout
    @param out: open file to write
    @param root: Root of the element tree
    @param indent: how much to indent each time
    @param verbose: set to true to get more debugging info
    '''
    if verbose: print 'building printFunc...\n'
    out.write('void '+className+'::print() {\n')
    out.write(indent+'std::cout << "class '+className+':\\n";\n')
    for category in root:
	if not category.tag:
	    continue # skip comments
	for optET in category:
	    if optET.tag!='opt': continue # Skip comments and other stuff
	    opt = Option(optET)

	    if verbose: print '    opt:', opt['name']
	    #var = optET.attrib['variable']
	    #varType = optET.attrib['type']
	    var = opt['variable']
	    varType = opt['type']

	    # FIX: handle if an array
	    if opt['repeatable']:
		out.write(indent+'std::cout << "'+indent+'" << "'+var+': array["<<'+var+'.size()<<"] ... \\n";\n')
		out.write(indent+'for (unsigned int i=0; i<'+opt['variable']+'.size();i++) {\n')
		out.write(indent*2+'std::cout << "'+2*indent+'" << '+var+'[i] << "\\n";\n')
		out.write(indent+'}\n')
	    else:
		out.write(indent+'std::cout << "'+indent+'" << "'+var+': " << '+var+' << "\\n";\n')

    out.write('} // '+className+'::print()\n\n')

######################################################################

def saveFunc(out,root,className,indent, verbose=True):
    '''
    Create a ::print c++ function to dump the class state to stdout
    @param out: open file to write
    @param root: Root of the element tree
    @param indent: how much to indent each time
    @param verbose: set to true to get more debugging info
    '''

    if verbose: print 'building saveFunc...\n'
    out.write('bool '+className+'::save(const std::string &filename) const {\n')
    #out.write('bool '+className+'::save(const std::string &filename, const '+className+'::Source &source) const {\n')
    #out.write(indent+'if (source!=All) std::cerr << "WARNING: source selection not implemented." << std::endl;\n')
    #print 'FIX: implement source selection.  Probably need to pass a vector of sources'
    out.write(indent+'std::ofstream o(filename.c_str(),std::ios::out);\n')
    out.write(indent+'o << "<?xml version=\\"1.0\\" encoding=\\"UTF-8\\"?>\\n";\n')
    out.write(indent+'o << "<argdump className=\\"'+className+'\\">\\n\\n";\n')
    for category in root:
	if not category.tag:
	    continue # skip comments
	for optET in category:
	    if optET.tag!='opt': continue # Skip comments and other stuff
	    opt = Option(optET)

	    if verbose: print '    opt:', opt['name']
	    var = opt['variable']
	    varType = opt['type']

	    out.write(indent+'o << "'+indent+'" << "<optvalues name=\\"'+opt['name']+'\\"\\n"\n')

	    out.write(indent+'  << "'+indent+'" << "variable=\\"'+opt['variable']+'\\"\\n"\n')
	    out.write(indent+'  << "'+indent+'" << "type=\\"'+opt['type']+'\\"\\n"\n')
	    out.write(indent+'  << "'+indent+'" << "repeatable=\\"'+str(opt['repeatable'])+'\\"\\n"\n')
	    #out.write(indent+'  << "'+indent+'" << "=\\"'+opt['']+'\\"\\n"\n')

	    #out.write(indent+'  << std::endl;\n')
	    out.write(indent+';\n')


	    # FIX: handle if an array
	    if opt['repeatable']:
		out.write(indent+'  '+'o << "'+indent+'>\\n";\n')
		out.write(indent+'for (unsigned int i=0; i<'+opt['variable']+'.size();i++) {\n')
		out.write(indent+'  '+'std::stringstream ostr;\n')
		out.write(indent+'  '+'ostr << '+opt['variable']+'[i];\n')
		out.write(indent+'  '+'o << "'+indent*2+'" << "<value>" << ostr.str() << "</value>\\n";\n')
		out.write(indent+'}\n')
		out.write(indent+'  '+'o << "'+indent+'" << "</optvalues>\\n\\n";\n')
		
	    else:
		out.write(indent+'{\n')
		out.write(indent+'  '+'std::stringstream ostr;\n')
		out.write(indent+'  '+'ostr << '+opt['variable']+';\n')
		out.write(indent+'  '+'o << "'+indent+'" << ">" << ostr.str() << "</optvalues>\\n\\n";\n')
		out.write(indent+'}\n')



    out.write(indent+'o << "</argdump>\\n";\n')
    out.write(indent+'return true;\n')
    out.write('} // '+className+'::save()\n\n')

######################################################################

def loadFunc(out,root,className,indent, verbose=True):
    '''
    Create a ::print c++ function to dump the class state to stdout
    @param out: open file to write
    @param root: Root of the element tree
    @param indent: how much to indent each time
    @param verbose: set to true to get more debugging info
    '''

    if verbose: print 'building loadFunc...\n'
    out.write('bool '+className+'::load(const std::string &filename) {\n')
    print 'FIX: implement loading using libxml2 selection.'

    for category in root:
	if not category.tag:
	    continue # skip comments
	for optET in category:
	    if optET.tag!='opt': continue # Skip comments and other stuff
	    opt = Option(optET)

	    if verbose: print '    opt:', opt['name']
	    var = opt['variable']
	    varType = opt['type']
	    print 'FIX: add code for each option!'

    out.write(indent+'assert(false && "Write code to read in a parameters file using libxml2");')

    out.write('} // '+className+'::load()\n\n')
     
######################################################################
def writeCxxSource(argsXmlFilename='genargs.xml', outFilename='Parameters.cxx'
		,gengetoptStructName='gengetopt_args_info'
		,gengetoptHeader='cmdline.h'
		,className='Parameters',indent='    '
		,headerFileName='Parameter.hxx', verbose=False):
    '''
    FIX: write docs here

    @param argsXmlFilename: FIX: write
    @param outFilename: FIX: write
    @param gengetoptStructName: FIX: write
    @param className: FIX: write
    @param indent: FIX: write
    '''
    from lxml import etree as ET
    root = ET.parse(argsXmlFilename).getroot()

    o = file(outFilename,'w')
    o.write('// WARNING: autogenerated C++ code.  DO NOT EDIT.  See genargs\n')
    o.write('/** @file *\n')
    o.write('    @brief Autogenerated cpp implementation code for parameters.\n')
    o.write('\n')
    o.write('*/\n')

    o.write('''
#include "'''+headerFileName+'''"
#include "'''+gengetoptHeader+'''"

// C std libraries includes
#include <cstdlib>

// C++ std library includes
#include <iostream>
#include <sstream> // string stream
#include <fstream>

''')

    writeLibxml2ReaderHelpers(o)
    writeLibxml2Reader(o,root,indent=indent,className=className,verbose=verbose)

    o.write('//////////////////////////////////////////////////////////////////////\n')
    o.write(className+'::'+className+'''(const bool useDefaults) {
'''+indent+'''if (useDefaults) init();
}

''')

    cmdlineConstructor(o,root,className,indent,gengetoptStructName)

    o.write('void '+className+'::init() {\n')

    for category in root:
	if not category.tag:
	    continue # skip comments
	for opt in category:
	    if opt.tag!='opt': continue # Skip comments and other stuff
	    if verbose: print '    opt:', opt.attrib['name']
	    writeOptInit(o,opt,indent)
    o.write('} // '+className+'::init()\n\n')


    printFunc(o,root,className,indent)

    
    saveFunc(o,root,className,indent)


    o.write('////////////////////\n')
    o.write('// TESTING CODE\n')
    o.write('////////////////////\n')
    o.write('#ifdef TESTING\n')
    o.write('int main (int argc, char *argv[]) {\n')
    o.write(indent+'std::cout << "Starting " << argv[0] << std::endl;\n')

    o.write(indent+'{\n')
    o.write(2*indent+className+' argsVar(argc,argv);\n')
    o.write(2*indent+'std::cout << "Calling print... " << std::endl;\n')
    #o.write(2*indent+'argsVar.print();\n')
    o.write(2*indent+'argsVar.save("'+className+'_dump.xml");\n')
    o.write(indent+'}\n')

    o.write(indent+'{\n')
    o.write(2*indent+className+' a(true);\n')
    o.write(2*indent+'a.intArray.push_back(2);\n')
    o.write(2*indent+'a.intArray.push_back(3);\n')
    o.write(2*indent+'a.intArray.push_back(4);\n')

    o.write(2*indent+'a.floatArray.push_back(5.15);\n')
    o.write(2*indent+'a.floatArray.push_back(6.16);\n')
    o.write(2*indent+'a.floatArray.push_back(7.17);\n')

    o.write(2*indent+'a.stringArray.push_back("eight");\n')
    o.write(2*indent+'a.stringArray.push_back("s9");\n')
    o.write(2*indent+'a.stringArray.push_back("s10");\n')

    o.write(2*indent+'std::cout << "Calling print... " << std::endl;\n')
    o.write(2*indent+'a.print();\n')
    o.write(2*indent+'a.save("'+className+'_dump.xml");\n')
    o.write(indent+'}\n')

    o.write(indent+'{\n')
    o.write(2*indent+className+' argsVar(false);\n')
    o.write(2*indent+'std::cout << "\\nargsVar.load('+className+'_dump.xml)\\n";\n')
    o.write(2*indent+'argsVar.load("'+className+'_dump.xml");\n')
    o.write(2*indent+'argsVar.print();\n')
    o.write(indent+'}\n')

    o.write(indent+'\n')
    o.write('}// main\n')
    o.write('#endif // TESTING\n')


######################################################################
if __name__=='__main__':
    from optparse import OptionParser
    myparser = OptionParser(usage="%prog [options]",
			    version="%prog "+__version__)
    
    myparser.add_option('-o','--output-base',dest='outputBasename',default='Parameters',
			help='Base filename of the gen C++ file to write [default: %default]')

    myparser.add_option('-a','--args-xml',dest='argsXmlFilename',default='genargs.xml',
			help='Name of file with the xml arg definitions [default: %default]')

    myparser.add_option('-c','--c-struct',dest='cStructName',default='gengetopt_args_info',
			type='string',
			help='Gengetopt structure name.  Requires gengetopt >=2.15.  '
			+ '[default: %default]'
			)

    
    myparser.add_option('-C','--class-name',dest='className',default='Parameters',
			help='C++ class name [default: %default]')

    myparser.add_option('--cxx-ext',dest='cxxExt',default='.cxx',type='string',
			help='C++ extention for sources (e.g. C, cpp, cxx)'
			+'[default: %default]'
			)
    myparser.add_option('--cxx-header-ext',dest='cxxHeaderExt',default='.hxx',type='string',
			help='C++ extention for header files (e.g. h, H, hpp, hxx)'
			+' [default: %default]'
			)

    myparser.add_option('--gengetopt-header',dest='gengetoptHeader',default='cmdline.c',type='string',
			help='Header file produced by gengetopt.  [default: %default]')

    myparser.add_option('-e','--example-compile',dest='exampleCompile',default=False,action='store_true',
			help='Show an example compile line for gcc.')

    myparser.add_option('-p','--public-members',dest='publicMembers',default=True, action='store_true',
			help='Make class members/variables public [default: %default]')
    myparser.add_option('-P','--private-members',dest='publicMembers',action='store_false',
			help='Make class members/variables private')

    myparser.add_option('-g','--get-set',dest='includeGetSet',default=True, action='store_true',
			help='Add get/set accessor methods for the members [default: %default]')
    myparser.add_option('-G','--without-get-set',dest='includeGetSet',action='store_false',
			help='Suppress get/set accessor methods for the members')


    (options,args) = myparser.parse_args()

    headerFileName=options.outputBasename+options.cxxHeaderExt

    writeCxxHeader(options.argsXmlFilename, headerFileName
		   ,gengetoptStructName=options.cStructName
		   ,className=options.className
		   ,publicMembers=options.publicMembers
		   ,includeGetSet=options.includeGetSet
		   )
    writeCxxSource(options.argsXmlFilename, options.outputBasename+options.cxxExt
		   ,gengetoptStructName=options.cStructName
		   ,gengetoptHeader=options.gengetoptHeader
		   ,className=options.className
		   ,headerFileName=headerFileName
		   )

    if options.exampleCompile:
	print 'Example command line to compile this code:'
	print '   g++ -g3 -Wall -Wimplicit -pedantic -W -Wredundant-decls -Werror -DTESTING -o ',
	# FIX: remove from the last . rather than this hack for the gengetopt obj
	print options.outputBasename,options.outputBasename+options.cxxExt,options.gengetoptHeader[:-2]+'.o'

    del headerFileName

